import java.util.*;

/**
 * Rules of The Resistance (www.proavaolon.com/rules)
 * The game requires between five and ten players.
 * Approximately one third of the players are randomly chosen as spies; the rest are members of the Resistance.
 * The spies know who the other spies are, but Resistance members do not.
 * The game consists of up to five missions. Each mission has a leader. The leader proposes a mission team of a certain size, which the group approves by public vote.
 * If the group does not approve the mission by a simple majority, leadership passes to the next player.
 * If the group cannot approve a mission team after five attempts, the spies win.
 * Once a mission team is chosen, it votes by secret ballot whether the mission succeeds or fails.
 * Resistance members will always vote for success, but spies have the option of voting for success or failure.
 * It usually only takes one spy on the mission team to sabotage a mission, but sometimes the fourth mission requires two.
 * If three missions succeed, the Resistance wins. If three fail, the spies win.
 * You are allowed to say anything, to any one, at any time -- as long as it is said publicly.
 */
public class Resistance {


    /**
     * Row i, Column j represents the team size for round i+1 and j + 5 number of players;
     * Negative indicates two negative votes required
     */
    public final int[][] missionTeamSizes = new int[][] {
            {2, 2, 2, 3, 3, 3},
            {3, 3, 3, 4, 4, 4},
            {2, 4, 3, 4, 4, 4},
            {3, 3, -4, -5, -5, -5},
            {3, 4, 4, 5, 5, 5}
    };

    public final int[] numberOfSpies = new int[] {2, 2, 3, 3, 3, 4};

    private int numPlayers, numSpies, roundNumber, leader;
    private int failedVotes, failedMissions, successfulMissions;
    private int[] spies;

    private ArrayList<Integer> missionTeam;

    public ArrayList<Player> players = new ArrayList<Player>();

    /**
     * Initialize a new game of certain number of players
     * @param numberOfPlayers
     */
    public Resistance(int numberOfPlayers, int[] presetSpies)
    {
        if(numberOfPlayers > 10 || numberOfPlayers < 5)
            numberOfPlayers = 5;
        numPlayers = numberOfPlayers;
        for(int i = 0; i < numberOfPlayers; i++)
            players.add(new Player(i+1));
        numSpies = numberOfSpies[numberOfPlayers-5];
        leader = 1;
        failedVotes = 0;
        failedMissions = 0;
        successfulMissions = 0;
        spies = presetSpies;
        roundNumber = 1;
        missionTeam = new ArrayList<Integer>();
    }

    public Resistance(int numberOfPlayers){
        this(numberOfPlayers, new int[0]);
        if(spies.length == 0)
            spies = randomlySelectSpies();
    }

    public void reportSpies() {
        String s = "The spies are: ";
        for (int spy : spies)
            s += Player.playerNames.get(spy) + ", ";
        ResistanceDriver.logger.log(s);
    }
    public void reportResistance() {
        String s = "The resistance are: ";
        for (int i = 1; i <= numPlayers; i++)
        {
            boolean isSpy = false;
            for(int spy : spies)
                if(spy == i)
                {
                    isSpy = true;
                    break;
                }
            if(!isSpy)
                s += Player.playerNames.get(i) + ", ";
        }
        ResistanceDriver.logger.log(s);
    }

    public void reportGameInfo() {
        ResistanceDriver.logger.log("Number of players: " + numPlayers);
        ResistanceDriver.logger.log("Number of spies: " + numSpies);
    }

    /**
     * Randomly select players to be spies.
     * @return
     */
    private int[] randomlySelectSpies() {
        int[] spies = new int[numSpies];
        ArrayList<Integer> selected = new ArrayList<Integer>();
        for (int i = 0; i < spies.length; i++) {
            spies[i] = (int)(Math.random()*numPlayers)+1;
            if(selected.contains(spies[i]))
                i--;
            else
                selected.add(spies[i]);
        }
        Arrays.sort(spies);
        return spies;
    }

    /**
     * Gets the team size required for the number of players and the round.
     * @return the team size
     */
    public int getMissionTeamSize(){
        return missionTeamSizes[roundNumber-1][numPlayers-5];
    }

    /**
     * Prompt the leader to select team members for a mission.
     */
    public void leaderSelectTeam(){
        players.get(leader-1).playerSelectTeam(missionTeam, numPlayers, Math.abs(getMissionTeamSize()));
    }

    /**
     * Players vote on proposed team (should be done simultaneously).
     */
    public boolean voteOnProposedTeam(){
        int yesVotes = 0;
        for(Player p : players)
            if(p.voteOnMissionTeam(missionTeam))
                yesVotes++;
        if (yesVotes > numPlayers/2)
        {
            ResistanceDriver.logger.log("Vote passes.");
            failedVotes = 0;
            return true;
        }
        else{
            ResistanceDriver.logger.log("Vote fails.");
            failedVotes += 1;
            return false;
        }
    }

    /**
     *Returns a string representing the state of a game.
     * @return
     */
    public String toString(){
        StringBuilder sb = new StringBuilder();
        sb.append("\nRound: " + roundNumber + "\nFailed Team Votes: " + failedVotes + "\n")
                .append("Leader: " + Player.playerNames.get(leader) + "\n")
                .append("Mission Score: Succeeded = " + successfulMissions + ", Failed = " + failedMissions + "\n");
        return sb.toString();
    }

    public void passLeader() {
        leader += 1;
        if (leader > numPlayers)
            leader = 1;

        ResistanceDriver.logger.log("Leader passes to " + Player.playerNames.get(leader));
    }

    public void goOnMission() {
        int fails = 0;
        for(int playerNumber : missionTeam)
            if(!players.get(playerNumber-1).goOnMission())
                fails += 1;
        if((getMissionTeamSize() > 0 && fails >= 1) || (getMissionTeamSize() < 0 && fails >= 2)) //2 fail votes needed in some cases
        {
            ResistanceDriver.logger.log("The mission fails with " + fails + " failure votes.");
            failedMissions += 1;
        }
        else{
            ResistanceDriver.logger.log("The mission passes with " + fails + " failure votes.");
            successfulMissions += 1;
        }
        roundNumber += 1;
    }

    public boolean isWinner() {
        if(failedVotes == 5)
        {
            ResistanceDriver.logger.log("5 failed votes, the resistance couldn't come to a consensus. The spies win.");
            return true;
        }
        if(failedMissions == 3)
        {
            ResistanceDriver.logger.log("3 failed missions, the spies win.");
            return true;
        }
        if(successfulMissions == 3)
        {
            ResistanceDriver.logger.log("3 successful mission, the resistance win.");
            return true;
        }
        return false;
    }

    public void deliberate() {
        ResistanceDriver.logger.log("Deliberation phase begins.");
        resetDeliberation();
        while(!allPlayersDoneDeliberating())
            for(Player p : players)
                if(!p.doneDeliberating)
                    p.deliberate();
        ResistanceDriver.logger.log("Deliberation phase ends.");
    }

    private void resetDeliberation() {
        for(Player p : players)
            p.doneDeliberating = false;
    }

    private boolean allPlayersDoneDeliberating(){
        for(Player p : players)
            if(!p.doneDeliberating)
                return false;
        return true;
    }


}
